package java.util.stream;

import java.util.concurrent.CountedCompleter;
import java.util.Spliterator;
import java.util.function.Supplier;
import java.util.function.Predicate;
import java.util.OptionalDouble;
import java.util.OptionalLong;
import java.util.OptionalInt;
import java.util.Optional;

final class FindOps
{
    private FindOps() {
    }
    
    public static <T> TerminalOp<T, Optional<T>> makeRef(boolean b) {
        return new FindOps.FindOp(b, StreamShape.REFERENCE, Optional.empty(), Optional::isPresent, FindOps.FindSink.OfRef::new);
    }
    
    public static TerminalOp<Integer, OptionalInt> makeInt(boolean b) {
        return new FindOps.FindOp(b, StreamShape.INT_VALUE, OptionalInt.empty(), OptionalInt::isPresent, FindOps.FindSink.OfInt::new);
    }
    
    public static TerminalOp<Long, OptionalLong> makeLong(boolean b) {
        return new FindOps.FindOp(b, StreamShape.LONG_VALUE, OptionalLong.empty(), OptionalLong::isPresent, FindOps.FindSink.OfLong::new);
    }
    
    public static TerminalOp<Double, OptionalDouble> makeDouble(boolean b) {
        return new FindOps.FindOp(b, StreamShape.DOUBLE_VALUE, OptionalDouble.empty(), OptionalDouble::isPresent, FindOps.FindSink.OfDouble::new);
    }
    
    private static final class FindOp<T, O> implements TerminalOp<T, O>
    {
        private final StreamShape shape;
        final boolean mustFindFirst;
        final O emptyValue;
        final Predicate<O> presentPredicate;
        final Supplier<TerminalSink<T, O>> sinkSupplier;
        
        FindOp(boolean mustFindFirst, StreamShape shape, O emptyValue, Predicate<O> presentPredicate, Supplier<TerminalSink<T, O>> sinkSupplier) {
            this.mustFindFirst = mustFindFirst;
            this.shape = shape;
            this.emptyValue = emptyValue;
            this.presentPredicate = presentPredicate;
            this.sinkSupplier = sinkSupplier;
        }
        
        @Override
        public int getOpFlags() {
            return StreamOpFlag.IS_SHORT_CIRCUIT | (this.mustFindFirst ? 0 : StreamOpFlag.NOT_ORDERED);
        }
        
        @Override
        public StreamShape inputShape() {
            return this.shape;
        }
        
        @Override
        public <S> O evaluateSequential(PipelineHelper<T> pipelineHelper, Spliterator<S> spliterator) {
            Object value = ((TerminalSink)pipelineHelper.wrapAndCopyInto((Sink)this.sinkSupplier.get(), spliterator)).get();
            return (value != null) ? value : this.emptyValue;
        }
        
        @Override
        public <P_IN> O evaluateParallel(PipelineHelper<T> pipelineHelper, Spliterator<P_IN> spliterator) {
            return new FindOps.FindTask(this, pipelineHelper, spliterator).invoke();
        }
    }
    
    private abstract static class FindSink<T, O> implements TerminalSink<T, O>
    {
        boolean hasValue;
        T value;
        
        FindSink() {
        }
        
        @Override
        public void accept(T value) {
            if (!this.hasValue) {
                this.hasValue = true;
                this.value = value;
            }
        }
        
        @Override
        public boolean cancellationRequested() {
            return this.hasValue;
        }
        
        static final class OfRef<T> extends FindOps.FindSink<T, Optional<T>>
        {
            @Override
            public Optional<T> get() {
                return this.hasValue ? Optional.of(this.value) : null;
            }
        }
        
        static final class OfInt extends FindOps.FindSink<Integer, OptionalInt> implements Sink.OfInt
        {
            @Override
            public void accept(int i) {
                super.accept(Integer.valueOf(i));
            }
            
            @Override
            public OptionalInt get() {
                return this.hasValue ? OptionalInt.of((int)this.value) : null;
            }
        }
        
        static final class OfLong extends FindOps.FindSink<Long, OptionalLong> implements Sink.OfLong
        {
            @Override
            public void accept(long l) {
                super.accept(Long.valueOf(l));
            }
            
            @Override
            public OptionalLong get() {
                return this.hasValue ? OptionalLong.of((long)this.value) : null;
            }
        }
        
        static final class OfDouble extends FindOps.FindSink<Double, OptionalDouble> implements Sink.OfDouble
        {
            @Override
            public void accept(double d) {
                super.accept(Double.valueOf(d));
            }
            
            @Override
            public OptionalDouble get() {
                return this.hasValue ? OptionalDouble.of((double)this.value) : null;
            }
        }
    }
    
    private static final class FindTask<P_IN, P_OUT, O> extends AbstractShortCircuitTask<P_IN, P_OUT, O, FindOps.FindTask<P_IN, P_OUT, O>>
    {
        private final FindOps.FindOp<P_OUT, O> op;
        
        FindTask(FindOps.FindOp<P_OUT, O> op, PipelineHelper<P_OUT> pipelineHelper, Spliterator<P_IN> spliterator) {
            super(pipelineHelper, spliterator);
            this.op = op;
        }
        
        FindTask(FindOps.FindTask<P_IN, P_OUT, O> findTask, Spliterator<P_IN> spliterator) {
            super(findTask, spliterator);
            this.op = findTask.op;
        }
        
        @Override
        protected FindOps.FindTask<P_IN, P_OUT, O> makeChild(Spliterator<P_IN> spliterator) {
            return new FindOps.FindTask<P_IN, P_OUT, O>(this, spliterator);
        }
        
        @Override
        protected O getEmptyResult() {
            return this.op.emptyValue;
        }
        
        private void foundResult(O o) {
            if (this.isLeftmostNode()) {
                this.shortCircuit(o);
            }
            else {
                this.cancelLaterNodes();
            }
        }
        
        @Override
        protected O doLeaf() {
            Object value = ((TerminalSink)this.helper.wrapAndCopyInto((Sink)this.op.sinkSupplier.get(), this.spliterator)).get();
            if (!this.op.mustFindFirst) {
                if (value != null) {
                    this.shortCircuit(value);
                }
                return null;
            }
            else {
                if (value != null) {
                    this.foundResult(value);
                    return value;
                }
                return null;
            }
        }
        
        @Override
        public void onCompletion(CountedCompleter<?> countedCompleter) {
            if (this.op.mustFindFirst) {
                FindOps.FindTask<Object, Object, Object> findTask = (FindOps.FindTask)this.leftChild;
                FindOps.FindTask<Object, Object, Object> findTask2 = null;
                while (findTask != findTask2) {
                    Object localResult = findTask.getLocalResult();
                    if (localResult != null && this.op.presentPredicate.test(localResult)) {
                        this.setLocalResult(localResult);
                        this.foundResult(localResult);
                        break;
                    }
                    findTask2 = findTask;
                    findTask = (FindOps.FindTask)this.rightChild;
                }
            }
            super.onCompletion(countedCompleter);
        }
    }
}
